//
//  FRC.swift
//  cviko-tv
//
//  Created by Martin Hruby on 17.04.18.
//  Copyright © 2018 Martin Hruby. All rights reserved.
//


import Foundation
import UIKit
import CoreData



// ---------------------------------------------------------------------
// Kombinuje TableView-DataSource a delegata FRC
// zpravy z FRC posila do tabulky
// ---------------------------------------------------------------------
// Je genericka trida, kde Type vsak musi byt odvozen od ManagedObject
class FRCTableDataSource<Type:CKSynced>: NSObject, UITableViewDataSource, NSFetchedResultsControllerDelegate
{
    // -----------------------------------------------------------------
    var _newInsertedIndpaths = [IndexPath]()
    
    // -----------------------------------------------------------------
    // DataSource
    func numberOfSections(in tableView: UITableView) -> Int {
        //
        if let _sec = FRC.sections {
            //
            return _sec.count
        }
        
        //
        return 0
    }
    
    // -----------------------------------------------------------------
    // DataSource
    func tableView(_ tableView: UITableView,
                   numberOfRowsInSection section: Int) -> Int
    {
        // predpokladam FRC.sections != nil
        return FRC.sections![section].numberOfObjects
    }
    
    // -----------------------------------------------------------------
    //
    func tableView(_ tableView: UITableView,
                   cellForRowAt indexPath: IndexPath) -> UITableViewCell
    {
        //
        let _cell = tableView.dequeueReusableCell(withIdentifier: cellID,
                                                  for: indexPath)
        
        // Volani konfiguracni procedury zpatky do ViewControlleru, ktery
        // vlastni tento FRC
        cfgProcedure(FRC.object(at: indexPath), _cell)
        
        //
        return _cell
    }
    
    // -----------------------------------------------------------------
    // "protokol" pro cfgProcedure
    typealias CFGProc = (Type, UITableViewCell) -> ()
    
    // -----------------------------------------------------------------
    // nastaveni z ViewControlleru
    let cellID: String
    let FRC: NSFetchedResultsController<Type>
    let cfgProcedure: CFGProc
    let tableView: UITableView
    
    // -----------------------------------------------------------------
    // Pokud chceme docasne "rozpojit" vazbu MOC->FRC->VC(Table), napr.
    // pri vetsich zmenach do objektu MOC
    // Alternativne: provadet v global vlakne (docasne MOC)
    func setPaused(_ pause: Bool) -> Bool {
        //
        if pause {
            //
            FRC.delegate = nil
        } else {
            //
            FRC.delegate = self
            do { try FRC.performFetch() } catch { return false }
            //
            tableView.reloadData()
        }
        
        //
        return true
    }
    
    // -----------------------------------------------------------------
    // Protokol NSFetchedResultsControllerDelegate
    // -----------------------------------------------------------------
    // aplikace zmen v poradi:
    // 1) controllerWillChangeContent
    // 2) dle potreb: didChange anObject: Any...
    // 3) controllerDidChangeContent
    func controllerWillChangeContent(_ controller: NSFetchedResultsController<NSFetchRequestResult>)
    {
        //
        print("Fetch updated \(FRC.fetchedObjects?.count) zaznamu")
        
        //
        tableView.beginUpdates()
    }
    
    //
    func controller(_ controller: NSFetchedResultsController<NSFetchRequestResult>, didChange sectionInfo: NSFetchedResultsSectionInfo, atSectionIndex sectionIndex: Int, for type: NSFetchedResultsChangeType)
    {
        switch type {
        case .insert:
            //
            tableView.insertSections(IndexSet([sectionIndex]), with: .fade)
        default:
            ()
        }
    }
    
    func controller(_ controller: NSFetchedResultsController<NSFetchRequestResult>, didChange anObject: Any, at indexPath: IndexPath?, for type: NSFetchedResultsChangeType, newIndexPath: IndexPath?)
    {
        switch type {
        case .insert:
            // tabulka pak bude sama volat dataSource pro obsah bunky
            tableView.insertRows(at: [newIndexPath!], with: .fade)
            //tableView.selectRow(at: newIndexPath!, animated: true,
            //                    scrollPosition: UITableViewScrollPosition.middle)
            _newInsertedIndpaths.append(newIndexPath!)
        case .delete:
            tableView.deleteRows(at: [indexPath!], with: .fade)
        case .update:
            //
            cfgProcedure(anObject as! Type, tableView.cellForRow(at: indexPath!)!)
        case .move:
            //
            cfgProcedure(anObject as! Type, tableView.cellForRow(at: indexPath!)!)
            //
            tableView.moveRow(at: indexPath!, to: newIndexPath!)
        }
    }
    
    func controllerDidChangeContent(_ controller: NSFetchedResultsController<NSFetchRequestResult>)
    {
        //
        tableView.endUpdates()
        
        //
        if let _fin = _newInsertedIndpaths.first {
            //
            tableView.selectRow(at: _fin, animated: true, scrollPosition: .middle)
            
            //
            _newInsertedIndpaths = []
        }
    }
    
    // -----------------------------------------------------------------
    // editace tabulky v editacnim modu: zpravy se posilaji dataSource
    // -----------------------------------------------------------------
    func tableView(_ tableView: UITableView,
                   commit editingStyle: UITableViewCellEditingStyle,
                   forRowAt indexPath: IndexPath)
    {
        //
        if editingStyle == .delete {
            //
            let _obj = FRC.object(at: indexPath)
            
            // registrace smazani objektu
            DataMAN.shared.MOC.delete(_obj)
            
            // jestli existuje ck-synchronizace objektu, pak
            // zkonstruuj ckrecord a zadej smazani na cloud
            if let _ = _obj.ckencoded, let _ckRecord = _obj.generateCK() {
                //
                DispatchQueue.global().async {
                    //
                    DataMAN.shared.syncDeletes(recIds: [_ckRecord.recordID])
                }
            }
        }
    }
    
    //
    func tableView(_ tableView: UITableView,
                   canEditRowAt indexPath: IndexPath) -> Bool
    {
        //
        return true
    }
    
    // -----------------------------------------------------------------
    // nastartovani FRC
    func startup() -> Bool {
        // stavam se NSFetchedResultsControllerDelegate
        FRC.delegate = self
        
        //
        do {
            // zde se provede iniciacni fetch, dale pak FRC sleduje
            // zmeny na MOC
            try FRC.performFetch()
        } catch {
            //
            return false
        }
        
        //
        return true
    }
    
    // -----------------------------------------------------------------
    //
    init(withRQ: NSFetchRequest<Type>,
         cellID: String,
         tableView: UITableView,
         cfg: @escaping CFGProc)
    {
        //
        self.FRC = NSFetchedResultsController(fetchRequest: withRQ, managedObjectContext: DataMAN.shared.MOC, sectionNameKeyPath: nil, cacheName: nil)
        
        //
        self.cfgProcedure = cfg
        self.tableView = tableView
        self.cellID = cellID
    }
}
